package Depot::Process;

# Process.pm -- author: Oliver Bossert
# Copyright (C) 2004 Oliver Bossert
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

use strict;

use Depot::Conf;
use Depot::Tools;
use Depot::Fee;


############################################################
sub new { # Generate a new Object
############################################################
  my $type = shift;
  my $class = ref($type) || $type;
  my $self = {};
  bless $self, $class;
  return $self;
}

############################################################
sub process { # Process the data
############################################################
  my $self = shift;
  my $data = shift;
  my $query = shift;

  # Add a new depot
  if ( defined($query->param('newdepot')) && $query->param('newdepot') ne "" ) {
    $data = $self->add_new_depot( $data, $query->param('newdepot'), $query );
  }

  # Close an open Position
  if ( defined($query->param('close')) && $query->param('close') == 1 ) {
    my $id = $query->param('id');
    $data = $self->close_position( $data, $id, $query );
  }

  # Add a new order
  if ( defined($query->param('neworder')) && $query->param('neworder') == 1 ) {
    $data = $self->new_order( $data, $query );
  }

  # Cancel an order
  if ( defined($query->param('cancel')) && $query->param('cancel') == 1 ) {
    $data = $self->cancel_order( $data, $query );
  }

  # Process the functions
  $data = $self->process_functions( $data, $query );

  # Add a comment
  if ( defined($query->param('comment')) ) {
    $data = $self->add_comment( $data, $query );
  }

  return $data;
}

############################################################
sub add_new_depot { # Process the data
############################################################
  my $self = shift;
  my $data = shift;
  my $name = shift;
  my $query = shift;
  my $filename = 0;

  foreach my $depot ( keys %{$data->{'depots'}} ) {
    if ( $depot =~ /(\d+)/ ) {
      $filename = $1 + 1 if ( $1 + 1 > $filename);
    }
  }
  $filename = "depot-" . sprintf("%06d", $filename) . ".xml";
  $data->{'depots'}{$filename}->{'depot'}->{'name'} = $name;
  $data->{'depots'}{$filename}->{'depot'}->{'cash'}->{'value'} = "10000";
  @{$data->{'depots'}{$filename}->{'depot'}->{'history'}->{'trade'}} = ();
  @{$data->{'depots'}{$filename}->{'depot'}->{'orderlist'}->{'order'}} = ();
  @{$data->{'depots'}{$filename}->{'depot'}->{'positions'}->{'position'}} = ();

  $query->param('depot', $filename);

  return $data;
}

############################################################
sub close_position { # Close an open pos.
############################################################
  my $self = shift;
  my $data = shift;
  my $id = shift;
  my $query = shift;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  my $nb = -1;
  foreach my $i ( 0 .. $#{$depot->{'positions'}->{'position'}} ) {
    if ( $depot->{'positions'}->{'position'}->[$i]->{'id'} == $id ) {
      $nb++;
      last;
    }
  }

  print STDERR ">>>>>>>> " . $nb . "\n";

  if ( $nb != -1 ) {
    my $close = {};

    $close->{'id'} = getid( \@{$depot->{'orderlist'}->{'order'}} );
    $close->{'date'} = timestamp();
    $close->{'wkn'} = $depot->{'positions'}->{'position'}->[$nb]->{'wkn'};
    $close->{'nb'} = $depot->{'positions'}->{'position'}->[$nb]->{'nb'} . "x";
    if ($depot->{'positions'}->{'position'}->[$nb]->{'type'} =~ /long/i ) {
      $close->{'type'} = 'short';
    } else {
      $close->{'type'} = 'long';
    }
    $close->{'if'}->{'name'} = "TRUE";
    @{$close->{'if'}->{'param'}} = ();

    push @{$depot->{'orderlist'}->{'order'}}, $close;
  }

  return $data;
}

############################################################
sub new_order { # Create a new order
############################################################
  my $self = shift;
  my $data = shift;
  my $query = shift;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  my $order = {};

  $order->{'id'} = getid( \@{$depot->{'orderlist'}->{'order'}} );
  $order->{'date'} = timestamp();
  $order->{'wkn'} = $query->param('wkn');
  $order->{'nb'} = $query->param('count');

  if ($query->param('buysell') =~ /sell/i) {
    $order->{'type'} = 'short';
  } else {
    $order->{'type'} = 'long';
  }

  my $func = $query->param('func');
  my @param;
  if ($func =~ /^(\w+?)\((.*?)\)$/) {
    $func = $1;
    @param = split /\s*,\s*/, $2;
  }

  $order->{'if'}->{'name'} = $func;
  @{$order->{'if'}->{'param'}} = ();

  foreach (my $i=0; $i<=$#param; $i++) {
    my $par = {};
    $par->{'name'} = "P" . ($i+1);
    $par->{'value'} = $param[$i];
    push @{$order->{'if'}->{'param'}}, $par;
  }

  push @{$depot->{'orderlist'}->{'order'}}, $order;

  return $data;
}

############################################################
sub add_comment { # Add a comment
############################################################
  my $self = shift;
  my $data = shift;
  my $query = shift;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  foreach my $h ( @{$depot->{'history'}->{'trade'}} ) {
    if ( $h->{'id'} == $query->param('id') ) {
      my $c = {};
      $c->{'id'} = getid( \@{$h->{'comment'}} );
      $c->{'value'} = $query->param('comment');
      push @{$h->{'comment'}}, $c;
      last;
    }
  }

  return $data;
}

############################################################
sub cancel_order { # Cancel an order
############################################################
  my $self = shift;
  my $data = shift;
  my $query = shift;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );
  return $data if ( !defined($query->param('id')) );

  my $id = -1;
  foreach my $o ( @{$depot->{'orderlist'}->{'order'}} ) {
    $id++;
    last if ($o->{'id'} == $query->param('id'));
  }

  splice @{$depot->{'orderlist'}->{'order'}}, $id, 1;

  return $data;
}


############################################################
sub process_functions { # Close an open pos.
############################################################
  my $self = shift;
  my $data = shift;
  my $query = shift;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  my $nr = -1;
  foreach my $o ( @{$depot->{'orderlist'}->{'order'}} ) {
    $nr++;
    my $value = getval( $o->{'wkn'} );
    next if ( !defined($value) );

    my $func = $o->{'if'}->{'name'};
    my @params = ();
    if ( defined(@{$o->{'if'}->{'param'}}) && $#{$o->{'if'}->{'param'}} != -1) {
      foreach my $p ( sort {$a->{'name'} cmp $b->{'name'} } @{$o->{'if'}->{'param'}} ) {
	push @params, $p->{'value'};
      }
    }
    my $execute = 0;

    # Test for the various functions
    if ( uc($func) eq "TRUE" ) {
      $execute = 1;
    } elsif ( uc($func) eq "ABOVE" ) {
      if ( $#params >= 0 && $params[0] < $value ) {
	$execute = 1;
      }
    } elsif ( uc($func) eq "BELOW" ) {
      if ( $#params >= 0 && $params[0] > $value ) {
	$execute = 1;
      }
    } elsif ( uc($func) eq "TRAILINGABOVE" ) {
      if ( $#params >= 1 && $params[0] + $params[1] < $value  ) {
	$execute = 1;
      } elsif ( $#params >= 1 && $params[0] > $value ) {
	$params[0] = $value;
      }
    } elsif ( uc($func) eq "TRAILINGBELOW" ) {
      if ( $#params >= 1 && $params[0] - $params[1] > $value  ) {
	$execute = 1;
      } elsif ( $#params >= 1 && $params[0] < $value ) {
	$params[0] = $value;
      }
    }

    # Re-set the parameters
    my $i=0;
    foreach my $p ( sort {$a->{'name'} cmp $b->{'name'} } @{$o->{'if'}->{'param'}} ) {
      $p->{'value'} = $params[$i];
      $i++;
    }

    # Try to execute an order
    if ( $execute == 1 ) {

      # Determine the price
      my $price;
      my $tcount;
      my $count = $o->{'nb'};
      if ($count =~ /(\d+)x$/i) {
	$tcount = $1;
	$price = $tcount * $value;
      } else {
	$tcount = int($count / $value);
	$price = $tcount * $value;
      }
      my $fee = fee_open( $tcount, $value );

      # Schauen ob noch Aktienpositionen offen sind
      # Ich gehe davon aus, dass man nicht gelichzeitig 
      # eine Long und Short Pos. hlt
      my $indepot = -1;
      foreach my $i ( 0 .. $#{$depot->{'positions'}->{'position'}} ) {
	if ( $depot->{'positions'}->{'position'}->[$i]->{'wkn'} == $o->{'wkn'} ) {
	  $indepot = $i;
	  last;
	}
      }

      my $comment = $func . "(" . join(", ", @params) . ") on " . timestamp();
      # There are no stocks of this type in the depot
      if ( $indepot == -1 ) {
	if ( $fee + ($tcount*$value) > $depot->{'cash'}->{'value'} ) {
	  $execute = 0;
	} else {
	  $data = $self->_buy_stocks( $data, $query, $o->{'wkn'}, $o->{'type'}, $tcount, $value, $fee );
	}
      # There are some stocks left in the depot
      } else {

	# Same type; add stocks
	if ( $depot->{'positions'}->{'position'}->[$indepot]->{'type'} eq $o->{'type'}) {
	  if ( $fee + ($tcount*$value) > $depot->{'cash'}->{'value'} ) {
	    $execute = 0;
	  } else {
	    $data = $self->_add_stocks( $data, $query, $o->{'wkn'}, $o->{'type'}, $tcount, $value, $fee, 
					$indepot, "Closed by " . $comment );
	  }
	}
	# Different type
	else {
	  # Equal amount
	  if ( abs($depot->{'positions'}->{'position'}->[$indepot]->{'nb'} - $tcount) == 0 ) {
	    $data = $self->_close_pos( $data, $query, $indepot, $fee, $depot->{'positions'}->{'position'}->[$indepot]->{'nb'},
					 "Closed by " . $comment );
	  } elsif ( $depot->{'positions'}->{'position'}->[$indepot]->{'nb'} > $tcount ) {
	    $data = $self->_close_pos( $data, $query, $indepot, "fee", $tcount, "Closed by " . $comment . " (reducing)");
	  } else {
	    my $all = $depot->{'positions'}->{'position'}->[$indepot]->{'nb'};
	    $data = $self->_close_pos( $data, $query, $indepot, "fee", $all, "Closed by " . $comment);
	    $tcount -= $all; 
	    $fee = fee_open( $tcount, $value );
	    if ( $fee + ($tcount*$value) > $depot->{'cash'}->{'value'} ) {
	      $data = $self->_buy_stocks( $data, $query, $o->{'wkn'}, $o->{'type'}, $tcount, $value, $fee );
	    }
	  }
	}

      }
    }

    # If executed, kill order
    if ( $execute == 1 ) {
      splice @{$depot->{'orderlist'}->{'order'}}, $nr, 1;
    }
  }

  return $data;
}






############################################################
sub _buy_stocks {
############################################################
  my $self = shift;
  my ($data, $query, $wkn, $type, $tcount, $value, $fee ) = @_;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  $depot->{'cash'}->{'value'} -= $fee + ($tcount * $value);
  my $pos = {};

  $pos->{'id'} = getid( \@{$depot->{'positions'}->{'position'}} );
  $pos->{'wkn'} = $wkn;
  $pos->{'type'} = $type;
  $pos->{'date'} = timestamp();
  $pos->{'nb'} = $tcount;
  $pos->{'open'} = $value;
  $pos->{'fee'} = $fee;

  push @{$depot->{'positions'}->{'position'}}, $pos;

  return $data;
}


############################################################
sub _add_stocks {
############################################################
  my $self = shift;
  my ($data, $query, $wkn, $type, $tcount, $value, $fee, $id, $comment ) = @_;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  my $addcnt = $depot->{'positions'}->{'position'}->[$id]->{'nb'};
  $data = $self->_close_pos( $data, $query, $id, 0, $addcnt, $comment . " (adding stocks)" );
  $data = $self->_buy_stocks( $data, $query, $wkn, $type, ($tcount+$addcnt), $value, $fee, $id );

  return $data;
}


############################################################
sub _close_pos {
############################################################
  my $self = shift;
  my ($data, $query, $id, $fee, $nb, $comment ) = @_;
  my $depot;
  $depot = $data->{'depots'}{$query->param('depot')}->{'depot'} if (defined($query->param('depot')));

  return $data if ( !defined($depot) | $depot eq "" );

  my $position;
  my $killed = 0;
  if ( $depot->{'positions'}->{'position'}->[$id]->{'nb'} > $nb ) {
    $position = $depot->{'positions'}->{'position'}->[$id];
    $depot->{'positions'}->{'position'}->[$id]->{'nb'} -= $nb;
  } else {
    $position = splice @{$depot->{'positions'}->{'position'}}, $id, 1;
    $killed = 1;
  }

  my $value = getval( $position->{'wkn'} );
  $fee = fee_close( $nb, $value ) if ($fee eq "fee");

  my $history = {};
  $history->{'id'} = getid( \@{$depot->{'history'}->{'trade'}} );
  $history->{'wkn'} = $position->{'wkn'};
  $history->{'nb'} = $nb;
  $history->{'type'} = $position->{'type'};

  $history->{'open'}->{'date'} = $position->{'date'};
  $history->{'open'}->{'value'} = $position->{'open'};

  $history->{'close'}->{'date'} = timestamp();
  $history->{'close'}->{'value'} = getval( $position->{'wkn'} );

  $history->{'fee'}->{'value'} = $fee;
  $history->{'fee'}->{'value'} += $position->{'fee'} if (defined($position->{'fee'}) & $killed == 1);

  @{$history->{comment}} = ();
  if ( defined( $comment ) ) {
    my $c = {};
    $c->{'id'} = "000000";
    $c->{'value'} = $comment;
    push @{$history->{comment}}, $c;
  }

  push @{$depot->{'history'}->{'trade'}}, $history;

  # Calculate the cash!
  if ( $position->{'type'} =~/long/i ) {
    $depot->{'cash'}->{'value'} += $nb * $value;
  } else {
    $depot->{'cash'}->{'value'} += $nb * $position->{'open'} + ( $nb * ( $position->{'open'} - $value ) );
  }

  return $data;
}



############################################################
sub getid{
############################################################
  my $refarr = shift;
  my $maxid = 0;
  foreach my $data ( @{$refarr} ) {
    if ( $data->{'id'} =~ /^\d+$/ && $data->{'id'} > $maxid ) {
      $maxid = $data->{'id'};
    }
  }
  return sprintf("%06d", ($maxid+1));
}



1;
